_ENV = AIContext

local function average_weapon_damage(parent_type)
  local winfo = WEAPONS[parent_type]
  if winfo.is_fighter
    local ftr_weapon,alt_weapon = get_fighter_weapon(parent_type, empire)
    if alt_weapon
      return max(WEAPONS[ftr_weapon].damage_fun.average, WEAPONS[alt_weapon].damage_fun.average)
    end
    return WEAPONS[ftr_weapon].damage_fun.average
  end
  return winfo.damage_fun.average
end

local function design_attack_value(design)
  local score = 0

  for i,mount,parent_type,mods in design.mounts()
    local part_info = PartInfo[parent_type]
    if part_info?.is_weapon
      local damage = average_weapon_damage(parent_type)
      if part_info.is_beam
        damage *= 0.67 --beam attenuation fudge factor
      end
      if part_info.is_fighter
        damage *= 0.17 --PD and flight time fudge factor. 
      elseif part_info.is_launcher
        damage *= 0.33 --PD fudge factor
      end

      score += damage * design.implied_weapon_n(i,mods,mount,parent_type)
    end
  end

  return score
end

local function estimated_boarding_crew_damage(hull_specs)
  local crew_hitchance = empire.tactics.boarding_bonus or 0
  return hull_specs.crew*(50 + crew_hitchance)/100 * 0.5 * 0.5
end

local boarding_systems = set("Boarding Module", "Assault Shuttles", "Transporter")

local function is_combat_boarder(design)
  for i,mount,parent_type,mods in design.mounts()
    if boarding_systems[parent_type]
      return parent_type
    end
  end
  return nil
end

local function design_defence_value(hull_specs)
  return hull_specs.shields + hull_specs.armor + hull_specs.hull*0.7
end

function design_combat_value(design)
  local specs = design_to_specs(design)

  local attack_value = design_attack_value(design)
  if is_combat_boarder(design)
    attack_value += 2*estimated_boarding_crew_damage(specs)
  end
  local defence_value = design_defence_value(specs)

  --print(design.design_name, attack_value, defence_value)

  return attack_value*(1+defence_value)
end
