_ENV = AIContext

local function find_best_defense_design(planet)
  --ships can be used for defence too, and depending on their cost, they may be a better choice.
  --so consider all designs here, not just stations

  -- a value between 0 and 1
  -- affects how much the AI values getting a ship sooner vs getting it later
  local patience = 0.85


  if( planet.num_orbitals+1>planet:max_orbitals() )
    return false
  end

  local designs in planet.empire

  local design_key, cost = get_max_key(designs, | ) function(key, design)
    if (design.fixed_defense or design.planet_info) and (planet.num_orbitals+1>planet:max_orbitals())
      return
    end

    local ship_value = design_combat_value(design)
    if ship_value == 0
      return
    end

    local build_time, projected_metal_income = estimate_build_time(key, planet)

    --Can't afford this. Don't build it unless we were already building it or we're out of minerals
    if empire.metal + build_time*projected_metal_income < 0 and key ~= planet.building_design
      return
    end

    if(design.fixed_defense) 
      ship_value *= 0.1
    end

    ship_value *= (patience^build_time)/build_time

    return ship_value
  end

  if not design_key
    return false
  end

  local ship = design_to_ship | designs[design_key]

  if(ship_power_score(ship) <= 0)
    return false
  end

  return design_key
end


local function can_build_defenses(planet)
  if not planet:cant_build_improvement 'Planetary Defenses'
    return true
  end

  local empire in planet

  return find_best_defense_design(planet)
end


function defense_need_metric(p)

  if not can_build_defenses(p)
    return
  end

  local threat, extra_enemy_fear in  threatened[p.star]!

  local threat = threat + extra_enemy_fear

  if (threat == 0) and not fully_developed(p)
    return
  end

  local mult = 10

  return  (p.defenses and 0 or 1)*mult + p.raw_hammer_production
end


local function build_defense(planet)

--  if(not planet.defenses and )
--    return "Planetary Defenses"
--  end

  print('building defense on',planet.name)

  if not planet:cant_build_improvement 'Planetary Defenses'
    order.set_building_improvement(planet, 'Planetary Defenses')
    return true
  end

  local design_key =   find_best_defense_design(planet)

  if design_key and not planet:cant_build_design(design_key)

    print('choose defense design', empire.designs[design_key].design_name, planet.name)

    order.set_building_design(planet,design_key)
    return true 
  end

end



function dispatch.defenses()

  local planet = best_planet(defense_need_metric)

  ai_debug_print( 'dispatch.defenses', nkeys(free_planets), planet?.name )


  if(not planet) 
    --console("no best planet")
    return 'pass' 
  end

  
  if build_defense(planet)
    free_planets[planet]=nil

    --console("set defense",planet)
  else

    --console("couldn't set a defense",planet)
    return 'pass'
  end
   
end

