_ENV = AIContext

depends "design_evaluation"

function init_production_desire.ships()

  maintinance_budget = empire:income()

-- print(..surplus)
-- production_desires.ships = at_war and 10 or 2

  total_ship_value = 0
  target_ship_value = target_navy_power()
  queued_ship_value = 0

  constructed_ships = {}

  for _,fleet ; empire.fleets
    for ship ; fleet.ships
      constructed_ships[ ship ]=true
      total_ship_value += combined_cost | ship
    end
  end


  design_to_specs = strong_memoize_1_to_1 | function(design)
    return design.update_hull_specs()
  end

  design_to_ship = strong_memoize_1_to_1 | function(design)
    local ship= EmulateRecorded.all_ships | design.genShip()
    ship.is_fixed = ship.drawer_table.fixed_defense
    return ship
  end


  update_production_desire.ships()


end

function get_best_infrastructure(planet)
  if planet.num_orbitals+1>planet:max_orbitals()
    -- this is a troublesome case -- the simplest fix is to just
    -- scrap a station

   -- local i,orbital = next | planet.orbitals

 
    local i, orbital = get_best_value(planet.orbitals, | ) function(b,a)
      return lexical_compare(
        a.construction_tier or 0, b.construction_tier or 0,
        combined_cost(a), combined_cost(b),
        a._ikey,b._ikey)
    end

    if not i
      error 'this is odd -- the planet really ought to have at least 1 orbital'
    end
  

    local _TRACE_INFO = planet.name

    print(..i,..orbital)

    print(..orbital.name)

    if orbital
        planet:scrap_orbital(i)
--       order.scrap_ship(planet,orbital)
    end
  end

  local design_key, design = get_best_value(empire.designs, | ) function(d0,d1)
  local s0 = design_to_specs(d0)
  local s1 = design_to_specs(d1)


    --if planet:cant_build_design(d0!._ikey) 
  --    print(v2s | d0)
--    end

  return lexical_compare(
    planet:cant_build_design(d0!._ikey) and 0 or 1, 
    planet:cant_build_design(d1!._ikey) and 0 or 1, 

    d0.planet_info and 0 or 1, d1.planet_info and 0 or 1, 
    s0.construction_tier or 0, s1.construction_tier or 0,
    combined_cost(s0), combined_cost(s1) )
  end

  return design_key, design
end

function build_best_infrastructure(planet)
  --errorf('build best infrastructure case at: %s',planet.name)

  local design_key = get_best_infrastructure(planet)

  if not design_key
     error 'thats odd, is the design set empty?' 
  end

  local cant = planet:cant_build_design(design_key)
  
  if cant
    print('TODO: handle this fallback case more appropriately -- unable to build due to ',cant)
    order.set_activity(planet,'Trade')
  else
    order.set_building_design(planet, design_key )
  end
  --print(..design_name)

end


function estimate_build_time(design_key, planet)
  if planet.hammer_production == 0
    return 999, empire.metal_production
  end

  local hammer_cost, metal_cost = planet:design_cost(design_key)

  --consider any pre-existing progress
  local hammers, building_metal = 0,0
  if planet.building_design == design_key
    hammers = planet.hammers
    building_metal = planet.building_metal
  else
    local config_name = planet.empire.designs[design_key].config_name
    for i,v ; planet.incomplete_projects?
      if v.config_name==config_name
        --can use this prior project
        hammers = v.hammers
        building_metal = v.building_metal
        break
      end
    end
  end

  hammer_cost = max(hammer_cost - hammers, 0)
  metal_cost = max(metal_cost - building_metal, 0)

  if hammer_cost == 0
    return 0, empire.metal_production
  end

  --consider hammers
  local turns_to_complete = max(hammer_cost - planet.overflow_hammers, 0)/planet.hammer_production

  --consider metal
  local projected_metal_use = metal_cost*(planet.hammer_production/hammer_cost)
  local projected_metal_income = empire.metal_production + planet.metal_use - projected_metal_use
  --print("estimated_build_time", empire.designs[design_key].design_name, ..projected_metal_use, ..projected_metal_income)

  if projected_metal_income < 0
    local turns_to_metal_stall = floor(-empire.metal/projected_metal_income)

    if turns_to_complete > turns_to_metal_stall

      local total_metal_use = 0
      for p ; empire.planets 
        if not p.no_colony and p ~= planet
          total_metal_use += p.metal_use
        end 
      end

      local projected_total_metal_use = total_metal_use + projected_metal_use
      local stalled_turns = turns_to_complete - turns_to_metal_stall
      turns_to_complete = turns_to_metal_stall + stalled_turns*(projected_total_metal_use/empire.total_metal_production)
      --print(..turns_to_metal_stall, ..stalled_turns, ..projected_total_metal_use, ..total_metal_production)
    end
  end

  if turns_to_complete < 0
    error("negative build time")
  end

  --print(turns_to_complete)

  return ceil(turns_to_complete), projected_metal_income
end

function value_design(design_key,planet)

  --local turns,cost = planet:turns_to_complete_design(design.design_name)
  local design = planet.empire.designs[design_key]
  local hammer_production in planet
  local specs = design_to_specs(design)
  local mod_cost = planet:design_cost( design_key )
  local eturns = mod_cost / hammer_production 

  if eturns > max_ship_production_turns
    return 0
  end

  -- only upgrade infrastructure if our ship building need is not
  -- too urgent.
  if max_ship_production_turns <= 15
    if planet : cant_build_design(design_key)
      return 0
    end
  end

  --local _TRACE_INFO = empire.name .. design.design_name

 -- print('specs',v2s|specs)

  if specs.scanner
    local mult = wants_scouts()
    if not mult
      return 0
    end
    return combined_cost(specs)*mult
  end

  --print('valuing',design.design_name)

  if (specs.initial_tanks or 0 ) > 0
    local mult = wants_transports()
    if not mult
      return 0
    end
    return combined_cost(specs) * mult
  end

  return combined_cost |specs

end


function choose_ship_production(planet)

  local remaining_budget = target_ship_value - queued_ship_value

  print(..remaining_budget)

  -- a value between 0 and 1
  -- affects how much the AI values getting a ship sooner vs getting it later
  local patience = 0.2^(1/max_ship_production_turns)


  local design_key, cost = get_max_key(empire.designs, | ) function(key, design)

    if design.fixed_defense or design.planet_info
      return
    end

    local specs = design_to_specs(design)
    local cost  = specs.hammer_cost + specs.metal_cost
    if cost > remaining_budget
      return
    end

    local ship_value = design_combat_value(design)
    if ship_value == 0
      return
    end

    local build_time, projected_metal_income = estimate_build_time(key, planet)

    --Can't afford this. Don't build it unless we were already building it or we're out of minerals
    if empire.metal + build_time*projected_metal_income < 0 and key ~= planet.building_design
      print('too expensive', design.design_name, ..build_time, ..projected_metal_income)
      return
    end

    if planet:cant_build_design(key) == 'construction_tier'
      local infrastructure = get_best_infrastructure(planet)
      build_time += estimate_build_time(infrastructure, planet)
    end

    ship_value *= (patience^build_time)/build_time

    if specs.initial_tanks or 0 > 0
      ship_value = max(ship_value, 100*wants_transports())
      ship_value /= (1 - 0.999*wants_transports())
    end

    if specs.scanner
      ship_value /= (1 - 0.999*wants_scouts())
    end

    print(empire.name, design.design_name, ship_value)
    return ship_value

  end

  if design_key
    print('choose design', empire.designs[design_key].design_name, planet.name, cost)
  end

  return design_key

end


function wants_scouts()
  if num_scoutable_stars == 0
    return 0
  end

  local scouts=0

  for ship ; constructed_ships
    if ship.scanner
      scouts++
    end
  end

  return saturate(1 - 4*scouts/num_scoutable_stars)
end

function wants_transports()
  local current_tanks, target_tanks = 0, 0
  for ship ; constructed_ships
    if ship.carrying_tanks > 0
      current_tanks += ship.carrying_tanks
    else
      target_tanks += ship_power_score(ship) / 400
    end

    --print(ship.name,ship.tanks)

  end

  target_tanks=min(target_tanks,10,max(2,nkeys(empire.planets)))

  --if current_tanks < target_tanks
  --  print(..current_tanks,..target_tanks)
  --end

  --print('YOYO', ..current_tanks,..target_tanks)

  return saturate(1 - current_tanks/target_tanks)
end